from __future__ import absolute_import
from __future__ import print_function
import argparse
import hashlib
import socket

import os
import paramiko
import six
import sys

import time

from cryptography.hazmat.primitives import serialization

from aetros import api
from aetros.utils import read_home_config, get_ssh_key_for_host, create_ssh_stream


class AuthenticateCommand:
    def __init__(self, logger):
        self.logger = logger
        self.client = None
        self.registered = False
        self.active = True

    def main(self, args):
        import aetros.const

        parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                                         prog=aetros.const.__prog__ + ' authenticate',
            description='Authenticates the machine with a new pair of SSH keys with a user account.')

        parsed_args = parser.parse_args(args)

        home_config = read_home_config()
        host = home_config['host']

        installed_key = get_ssh_key_for_host(host)
        key_exists_and_valid = False
        if installed_key:
            try:
                create_ssh_stream(home_config, exit_on_failure=False)
                key_exists_and_valid = True
            except Exception: pass

        if key_exists_and_valid:
            choice = six.moves.input("You have already configured a valid SSH (ssk_key: "+installed_key+") "
                                     "for "+host+".\nWant to create a new key? (y/N): ").lower()
            if choice != 'y' and choice != 'yes':
                print("Aborted.")
                sys.exit(1)

        ssh_key = paramiko.RSAKey.generate(4096)
        ssh_key_private = ssh_key.key.private_bytes(
            serialization.Encoding.PEM, serialization.PrivateFormat.TraditionalOpenSSL, serialization.NoEncryption()
        ).decode()
        ssh_key_public = 'rsa ' + ssh_key.get_base64()

        fingerprint = hashlib.md5(ssh_key.__str__()).hexdigest()
        fingerprint = ':'.join(a + b for a, b in zip(fingerprint[::2], fingerprint[1::2]))

        token = api.http_request('machine-token', None, {
            'host': socket.getfqdn(),
            'key': ssh_key_public
        })

        print("Open following link and login to confirm this machine's SSH key in your account.")
        print("Public Key Fingerprint: MD5:" + fingerprint)
        print("\n   https://" + host + "/confirm-machine/" + token)
        print("\nWaiting for confirmation ...")

        while True:
            time.sleep(3)
            response = api.http_request('machine-token/authorized?id=' + token, method='post')
            if response['status'] == 'confirmed':
                print("\n" + response['username'] + ' confirmed the public key. Test with "aetros id" or "ssh git@'+host+'".')
                private_key_path = os.path.expanduser('~/.ssh/aetros_' + response['username']+'_rsa')
                public_key_path = os.path.expanduser('~/.ssh/aetros_' + response['username']+'_rsa.pub')

                if not os.path.exists(os.path.dirname(private_key_path)):
                    os.makedirs(os.path.dirname(private_key_path))

                with open(private_key_path, 'w') as f:
                    f.write(ssh_key_private)

                with open(public_key_path, 'w') as f:
                    f.write(ssh_key_public)

                os.chmod(private_key_path, 0o600)
                os.chmod(public_key_path, 0o600)

                ssh_config_path = os.path.expanduser('~/.ssh/config')

                if not os.path.exists(os.path.dirname(ssh_config_path)):
                    os.makedirs(os.path.dirname(ssh_config_path))

                host_section = 'host '+host+'\n'
                identity_section = '    IdentityFile ~/.ssh/aetros_' + response['username']+'_rsa\n'

                if os.path.exists(ssh_config_path):
                    import re
                    regex = re.compile(r"^host\s+" + re.escape(host)+'\s*', re.IGNORECASE | re.MULTILINE)
                    with open(ssh_config_path, 'r+') as f:
                        config = f.read()

                        if regex.match(config):
                            config = regex.sub(host_section + identity_section, config, 1)
                        else:
                            config = host_section + identity_section + config

                        f.seek(0)
                        f.write(config)
                else:
                    with open(ssh_config_path, 'w') as f:
                        f.write(host_section + identity_section)

                print("Private key " + private_key_path + " installed in ~/.ssh/config for "+host+".\n")
                user = api.user()
                print("Key installed of account %s (%s)." % (user['username'], user['name']))
                sys.exit(0)
            if response['status'] == 'expired':
                print("Token expired.")
                sys.exit(1)





from __future__ import print_function, division
from __future__ import absolute_import

import time
import json
import os
import subprocess
import sys
import six

from aetros.logger import GeneralLogger
from aetros.utils import unpack_full_job_id, read_home_config, flatten_parameters, get_ssh_key_for_host
from aetros.const import JOB_STATUS
from .backend import JobBackend
from .Trainer import Trainer

start_time = time.time()


class GitCommandException(Exception):
    cmd = None


def start(logger, full_id, fetch=True, env=None, volumes=None, gpu_devices=None):
    """
    Starts the training process with all logging of a job_id
    """

    owner, name, id = unpack_full_job_id(full_id)

    if isinstance(sys.stdout, GeneralLogger):
        # we don't want to have stuff written to stdout before in job's log
        sys.stdout.clear_buffer()

    job_backend = JobBackend(model_name=owner + '/' + name)

    if fetch:
        job_backend.fetch(id)

    job_backend.restart(id)
    job_backend.start()
    job_backend.set_status('PREPARE')
    job_backend.monitoring_thread.handle_max_time = False

    start_command(logger, job_backend, env, volumes, gpu_devices=gpu_devices)


def start_command(logger, job_backend, env=None, volumes=None, gpu_devices=None):
    work_tree = job_backend.git.work_tree
    home_config = read_home_config()

    if not env:
        env = {}

    if 'PYTHONPATH' not in env:
        env['PYTHONPATH'] = os.getenv('PYTHONPATH', '')

    env['PYTHONPATH'] += ':' + os.getcwd()
    env['AETROS_MODEL_NAME'] = job_backend.model_name
    env['AETROS_JOB_ID'] = str(job_backend.job_id)
    env['DEBUG'] = os.getenv('DEBUG', '')
    env['AETROS_ATTY'] = '1'
    env['AETROS_GIT'] = job_backend.git.get_base_command()

    if os.getenv('AETROS_SSH_KEY_BASE64'):
        env['AETROS_SSH_KEY_BASE64'] = os.getenv('AETROS_SSH_KEY_BASE64')
    elif get_ssh_key_for_host(home_config['host']):
        # we need to read the key into env so the docker container can connect to AETROS
        env['AETROS_SSH_KEY_BASE64'] = open(get_ssh_key_for_host(home_config['host']), 'r').read()

    job_config = job_backend.job['config']

    if 'command' not in job_config :
        job_backend.fail('No "command" given. See Configuration section in the documentation.')

    command = job_config['command']
    image = job_config['image']

    if job_backend.is_simple_model():
        if image:
            command = ['python']
        else:
            command = [sys.executable]
        command += ['-m', 'aetros', 'start-simple', job_backend.model_name + '/' + job_backend.job_id]

    # replace {{batch_size}} parameters
    if isinstance(job_config['parameters'], dict):
        for key, value in six.iteritems(flatten_parameters(job_config['parameters'])):
            if isinstance(command, list):
                for pos, v in enumerate(command):
                    if isinstance(command[pos], six.string_types):
                        command[pos] = command[pos].replace('{{' + key + '}}', json.dumps(value))
            else:
                command = command.replace('{{' + key + '}}', json.dumps(value))

    logger.info("Switch working directory to " + work_tree)
    os.chdir(job_backend.git.work_tree)

    docker_image_built = False
    if job_config['dockerfile'] or job_config['install']:
        dockerfile = job_config['dockerfile']
        if isinstance(dockerfile, six.string_types) and os.path.exists(dockerfile):
            pass
        else:
            if isinstance(dockerfile, six.string_types):
                dockerfile_content = dockerfile
            elif isinstance(dockerfile, list) and len(dockerfile) > 0:
                dockerfile_content = "\n".join(dockerfile)
            else:
                if image is None:
                    job_backend.fail("Image name missing, needed by `install` in aetros.yml")
                dockerfile_content = 'FROM ' + image + '\nRUN '

                if isinstance(job_config['install'], list):
                    dockerfile_content += '\n RUN '.join(job_config['install'])
                else:
                    dockerfile_content += job_config['install']

            dockerfile_content = '# CREATED BY AETROS because of "install" or "dockerfile" config in aetros.yml.\n' \
                                 + dockerfile_content

            with open('Dockerfile.aetros', 'w') as f:
                f.write(dockerfile_content)

            dockerfile = 'Dockerfile.aetros'
            job_backend.commit_file('Dockerfile.aetros')

        job_backend.set_system_info('image/dockerfile', dockerfile)
        docker_build = [
            home_config['docker'],
            'build',
            '-t', job_backend.model_name,
            '-f', dockerfile,
            '.',
        ]

        logger.info("Prepare docker image: $ " + (' '.join(docker_build)))
        job_backend.set_status('IMAGE BUILD')
        p = execute_command(args=docker_build, bufsize=1, stderr=subprocess.PIPE, stdout=subprocess.PIPE)

        if p.returncode:
            job_backend.fail('Image build error')
            sys.exit(p.returncode)

        docker_image_built = True
        image = job_backend.model_name

    docker_command = None
    if image is not None:
        if not docker_image_built:
            logger.info("Pull docker image: $ " + image)
            job_backend.set_status('IMAGE PULL')
            execute_command(args=[home_config['docker'], 'pull', image], bufsize=1,
                stderr=subprocess.PIPE, stdout=subprocess.PIPE)

        inspections = execute_command_stdout([home_config['docker'], 'inspect', image])
        inspections = json.loads(inspections.decode('utf-8'))
        if inspections:
            inspection = inspections[0]
            with job_backend.git.batch_commit('Docker image'):
                job_backend.set_system_info('image/id', inspection['Id'])
                job_backend.set_system_info('image/docker_version', inspection['DockerVersion'])
                job_backend.set_system_info('image/created', inspection['Created'])
                job_backend.set_system_info('image/container', inspection['Container'])
                job_backend.set_system_info('image/architecture', inspection['Architecture'])
                job_backend.set_system_info('image/os', inspection['Os'])
                job_backend.set_system_info('image/size', inspection['Size'])
                job_backend.set_system_info('image/rootfs', inspection['RootFS'])

        # make sure old container is removed
        subprocess.Popen([home_config['docker'], 'rm', job_backend.job_id], stderr=subprocess.PIPE).wait()

        docker_command = [home_config['docker'], 'run', '-t', '--name', job_backend.job_id]
        docker_command += home_config['docker_options']

        env['AETROS_GIT_WORK_DIR'] = '/job'
        docker_command += ['--mount', 'type=bind,source='+job_backend.git.work_tree+',destination=/job']

        env['AETROS_STORAGE_DIR'] = '/aetros'
        docker_command += ['--mount', 'type=bind,source='+job_backend.git.git_path+',destination='+'/aetros/' + job_backend.model_name + '.git']

        home_config_path = os.path.expanduser('~/aetros.yml')
        if os.path.exists(home_config_path):
            env['AETROS_HOME_CONFIG_FILE'] = '/aetros/aetros.yml'
            docker_command += ['--mount', 'type=bind,source='+home_config_path+',destination='+'/aetros/aetros.yml']

        docker_command += ['-w', '/job']

        # make sure the docker command receives all environment variables
        for k in six.iterkeys(env):
            docker_command += ['-e', k]

        if volumes:
            for volume in volumes:
                docker_command += ['-v', volume]

        if 'resources' in job_backend.job:
            assigned_resources = job_backend.job['resources']

            cpus = 1
            if 'cpu' in assigned_resources and assigned_resources['cpu']:
                cpus = assigned_resources['cpu']
            docker_command += ['--cpus', str(cpus)]

            memory = 1
            if 'memory' in assigned_resources and assigned_resources['memory']:
                memory = assigned_resources['memory']

            docker_command += ['--memory', str(memory * 1024 * 1024 * 1024)]

        if gpu_devices and (sys.platform == "linux" or sys.platform == "linux2"):
            # only supported on linux
            docker_command += ['--runtime', 'nvidia']
            docker_command += ['-e', 'NVIDIA_VISIBLE_DEVICES=' + (','.join(gpu_devices))]
            # support nvidia-docker1 as well
            # docker_command += ['--device', '/dev/nvidia1']

        docker_command.append(image)

        # since linux doesnt handle SIGINT when pid=1 process has no signal listener,
        # we need to make sure, we attached one to the pid=1 process
        trap = 'trapIt () { "$@"& pid="$!"; trap "kill -INT $pid" INT TERM; ' \
               'while kill -0 $pid > /dev/null 2>&1; do wait $pid; ec="$?"; done; exit $ec;};'

        if isinstance(command, list):
            command = ' '.join(command)

        docker_command += ['sh', '-c', trap + 'trapIt ' + command]
        command = docker_command

    job_backend.set_system_info('image/name', str(image))

    if not isinstance(command, list):
        command = ['sh', '-c', command]

    p = None
    exited = False
    wait_stdout = None
    wait_stderr = None
    try:
        job_backend.set_status('STARTED')
        logger.warning("$ %s " % (' '.join([json.dumps(a) for a in command])))

        command_env = os.environ.copy()
        command_env.update(env)

        # make sure maxTime limitation is correctly calculated
        job_backend.monitoring_thread.handle_max_time = True
        job_backend.monitoring_thread.handle_max_time_time = time.time()

        # Since JobBackend sends SIGINT to its current process group, it sends also to its parents when same pg.
        # We need to change the process group of the process, so this won't happen.
        # If we don't this, the master process receives the SIGINT as well.
        kwargs = {}
        if os.name == 'nt':
            kwargs['creationflags'] = subprocess.CREATE_NEW_PROCESS_GROUP
        else:
            kwargs['preexec_fn'] = os.setsid

        p = subprocess.Popen(args=command,
            bufsize=1, stderr=subprocess.PIPE, stdout=subprocess.PIPE,
            env=command_env, **kwargs)
        wait_stdout = sys.stdout.attach(p.stdout)
        wait_stderr = sys.stderr.attach(p.stderr)

        p.wait()
        wait_stdout()
        wait_stderr()

        exited = True

        sys.exit(p.returncode)
    except SystemExit:
        # We can not send a SIGINT to the child process
        # as we don't know whether it received it already (pressing CTRL+C) or not (sending SIGINT to this process only
        # instead of to the group), so we assume it received it. A second signal would force the exit.
        # sys.__stdout__.write("SystemExit with " + str(p.returncode) + ', exited: ' + str(exited) + ", early: "+str(job_backend.in_early_stop)+"\n")

        # make sure the process dies
        if docker_command:
            # docker run does not proxy INT signals to the docker-engine,
            # so we need to do it on our own directly.
            subprocess.Popen([home_config['docker'], 'kill', '--signal', 'INT', job_backend.job_id],
                stderr=subprocess.PIPE, stdout=subprocess.PIPE).wait()
            subprocess.Popen([home_config['docker'], 'wait', job_backend.job_id], stdout=subprocess.PIPE).wait()
        elif not exited and p and p.poll() is None:
            p.kill() # sends SIGINT
            p.wait()

        if exited:
            if p.returncode == 0:
                job_backend.stop(progress=JOB_STATUS.PROGRESS_STATUS_DONE)
            elif p.returncode == 1:
                job_backend.stop(progress=JOB_STATUS.PROGRESS_STATUS_ABORTED)
            else:
                job_backend.stop(progress=JOB_STATUS.PROGRESS_STATUS_FAILED)
        else:
            # master received SIGINT before the actual command exited.
            if not job_backend.in_early_stop:
                # master did not receive early_stop signal (maxTime limitation)
                # if not, the master received a stop signal by server or by hand (ctrl+c), so mark as aborted
                job_backend.abort()
            else:
                # let the on_shutdown listener handle the rest
                pass

                # only
                #     # check if there was a JobBackend in the command
                #     # if so, we do not add any further stuff to the git
                #     last_progress = job_backend.git.contents('aetros/job/status/progress.json')
                #     if last_progress is not None and int(last_progress) > 2:
                #         # make sure, we do not overwrite stuff written inside the sub-process by using our own
                #         # on_shutdown listener.
                #         job_backend.stop()


def execute_command_stdout(command, input=None):
    p = subprocess.Popen(command, bufsize=1, stderr=subprocess.PIPE, stdout=subprocess.PIPE)
    out, err = p.communicate(input)

    if p.returncode:
        sys.stderr.write(out)
        sys.stderr.write(err)
        raise Exception('Could not execute command: ' + str(command))

    return out


def execute_command(**kwargs):
    p = subprocess.Popen(**kwargs)
    wait_stdout = sys.stdout.attach(p.stdout)
    wait_stderr = sys.stderr.attach(p.stderr)

    p.wait()
    wait_stdout()
    wait_stderr()

    return p


def git_execute(logger, repo_path, args):
    args = ['git', '--git-dir', repo_path + '/.git', '--work-tree', repo_path] + args
    logger.info("$ %s" % (' '.join(args), ))

    p = execute_command(args=args, bufsize=1, stderr=subprocess.PIPE, stdout=subprocess.PIPE)

    if p.returncode != 0:
        exception = GitCommandException("Git command returned not 0. " + (' '.join(args)))
        exception.cmd = (' '.join(args))
        raise exception


def start_keras(logger, job_backend):
    if 'KERAS_BACKEND' not in os.environ:
        os.environ['KERAS_BACKEND'] = 'tensorflow'

    from . import keras_model_utils

    # we need to import keras here, so we know which backend is used (and whether GPU is used)
    os.chdir(job_backend.git.work_tree)
    logger.debug("Start simple model")

    # we use the source from the job commit directly
    with job_backend.git.batch_commit('Git Version'):
        job_backend.set_system_info('git_remote_url', job_backend.git.get_remote_url('origin'))
        job_backend.set_system_info('git_version', job_backend.git.job_id)

    # all our shapes are Tensorflow schema. (height, width, channels)
    import keras.backend
    if hasattr(keras.backend, 'set_image_dim_ordering'):
        keras.backend.set_image_dim_ordering('tf')

    if hasattr(keras.backend, 'set_image_data_format'):
        keras.backend.set_image_data_format('channels_last')

    from .KerasCallback import KerasCallback
    trainer = Trainer(job_backend)
    keras_logger = KerasCallback(job_backend, job_backend.logger)

    job_backend.progress(0, job_backend.job['config']['epochs'])

    logger.info("Start training")
    keras_model_utils.job_start(job_backend, trainer, keras_logger)

    job_backend.done()


from __future__ import absolute_import
from __future__ import print_function

import json
import os
import sys
import tempfile
import urllib
from PIL import Image
import numpy as np

class JobModel:
    """
    :type job : dict
    """
    def __init__(self, id, job, storage_dir):
        self.id = id
        self.job = job
        self.storage_dir = storage_dir
        self._layers = None
        self._datasets = None

    @property
    def config(self):
        return self.job['config']

    @property
    def model_settings(self):
        return self.job['modelSettings']

    @property
    def insights_enabled(self):
        return 'insights' in self.config and self.config['insights']

    def get_model_node(self, name):
        if not isinstance(self.layers, list):
            return None

        for nodes in self.layers:
            if isinstance(nodes, list):
                for node in nodes:
                    if 'varName' in node and node['varName'] == name:
                        return node
                    if node['id'] == name or ('name' in node and node['name'] == name):
                        return node

        return None

    def get_batch_size(self):
        return self.job['config']['batchSize']

    def get_model_h5_path(self):
        return os.path.normpath(os.getcwd() + '/aetros/model.h5')

    def get_layers_path(self):
        return os.path.normpath(os.getcwd() + '/aetros/layer.json')

    def get_dataset_dir(self):
        return os.path.normpath(os.getcwd() + '/aetros/dataset')

    def get_dataset_downloads_dir(self, dataset):
        return os.path.normpath(self.storage_dir + '/aetros/dataset/%s/datasets_downloads' % (dataset['id'],))

    def get_weights_filepath_latest(self):
        return os.path.normpath(os.getcwd() + '/aetros/weights/latest.hdf5')

    def get_weights_filepath_best(self):
        return os.path.normpath(os.getcwd() + '/aetros/weights/best.hdf5')

    def get_input_names(self):
        names = []
        for node in self.layer[0]:
            names.append(node['varName'])

        return names

    def get_input_dataset_names(self):
        names = []
        for node in self.layers[0]:
            names.append(node['datasetId'])

        return names

    def set_input_shape(self, trainer):

        trainer.input_shape = {}

        for node in self.layers[0]:
            size = (int(node['width']), (node['height']))
            if node['inputType'] == 'image':
                shape = (size[0], size[1], 1)
            elif node['inputType'] == 'image_rgb':
                shape = (size[0], size[1], 3)
            else:
                shape = (size[0] * size[1],)

            if 'varName' in node:
                trainer.input_shape[node['varName']] = shape
            else:
                # older models
                trainer.input_shape = shape

    def is_python_model(self):
        return 'type' in self.job['config'] and self.job['config']['type'] != 'simple'

    def is_keras_model(self):
        return 'type' in self.job['config'] and self.job['config']['type'] == 'simple'

    def get_built_model(self, trainer):
        # its built with model designer
        model_provider = self.get_model_provider()
        model = model_provider.get_model(trainer)

        loss = model_provider.get_loss(trainer)
        optimizer = model_provider.get_optimizer(trainer)

        model_provider.compile(trainer, model, loss, optimizer)

        return model

    def predict(self, trainer, input):
        prediction = trainer.model.predict(input)

        top5 = np.argsort(-prediction[0])[:5]

        result = []
        for i in top5:
            result.append({
                'class': self.get_dataset_class_label(trainer, self.get_first_output_layer(), i),
                'prediction': float(prediction[0][i])
            })

        return result

    def encode_input_to_input_node(self, input, input_node):

        return input

    def convert_file_to_input_node(self, file_path, input_node=None):
        if input_node is None:
            input_node = self.get_input_node(0)

        size = (int(input_node['width']), int(input_node['height']))

        if 'http://' in file_path or 'https://' in file_path:
            local_path = tempfile.mktemp()
            print("Download input ...")
            f = open(local_path, 'wb')
            f.write(urllib.urlopen(file_path).read())
            f.close()
        else:
            local_path = file_path

        if input_node['inputType'] == 'list':
            raise Exception("List input not yet available")
        else:
            try:
                image = Image.open(local_path)
            except Exception:
                print(("Could not open %s" % (local_path,)))
                return []

            image = image.resize(size, Image.ANTIALIAS)

            from keras.preprocessing.image import img_to_array

            image = self.convert_image_to_node(image, input_node)

            return image

    def convert_image_to_node(self, image, input_node=None):
        from keras.preprocessing.image import img_to_array

        if input_node is None:
            input_node = self.get_input_node(0)

        if input_node['inputType'] == 'image':
            image = image.convert("L")
            image = img_to_array(image)

        elif input_node['inputType'] == 'image_bgr':
            image = image.convert("RGB")
            image = np.asarray(image, dtype='float32')
            image = image[:, :, ::-1].copy()
            image = img_to_array(image)
        else:
            image = image.convert("RGB")
            image = img_to_array(image)

        if 'imageScale' not in input_node:
            input_node['imageScale'] = 255

        if float(input_node['imageScale']) > 0:
            image = image / float(input_node['imageScale'])

        return image

    def get_model_provider(self):

        sys.path.append('./')

        model = __import__('model', '')
        print("Imported model script from %s " % (os.path.abspath('./model.py'),))
        sys.path.pop()

        return model
    
    @property
    def layers(self):
        if self._layers is None:
            with open(self.get_layers_path(), 'r') as f:
                self._layers = json.loads(f.read())

        return self._layers

    @property
    def datasets(self):
        if self._datasets is None:
            self._datasets = {}

            def read_dataset(id):
                dataset_json_path = os.path.normpath(self.get_dataset_dir() + '/' + id + '.json')
                with open(dataset_json_path, 'r') as f:
                    self._datasets[id] = json.loads(f.read())

            for owner in os.listdir(self.get_dataset_dir()):
                path = os.path.normpath(self.get_dataset_dir() + '/' + owner + '/dataset/')
                for name in os.listdir(path):
                    if name.endswith('.json'):
                        read_dataset(owner + '/dataset/' + name[:-len('.json')])

                    if os.path.isdir(path + name):
                        for sub in os.listdir(path + name):
                            if sub.endswith('.json'):
                                read_dataset(owner + '/dataset/' + name + '/' + sub[:-len('.json')])

        return self._datasets

    def get_datasets(self, trainer):
        datasets = {}

        from aetros.utils import get_option
        from .auto_dataset import get_images, read_images_keras_generator, read_images_in_memory

        # load placeholder, auto data
        config = self.job['config']
        for layer in self.layers[0]:
            if 'datasetId' in layer and layer['datasetId']:

                if layer['datasetId'] not in self.datasets:
                    raise Exception('Dataset %s not found in datasets %s' % (layer['datasetId'], json.dumps(self.datasets.keys())))

                dataset = self.datasets[layer['datasetId']]
                if not dataset:
                    raise Exception('Dataset of id %s does not exists. Available %s' % (
                    layer['datasetId'], ','.join(list(self.datasets.keys()))))

                if dataset['type'] == 'images_upload' or dataset['type'] == 'images_search':

                    connected_to_layer = self.get_connected_layer(self.layers, layer)
                    if connected_to_layer is None:
                        # this input is not in use, so we dont need to calculate its dataset
                        continue

                    datasets[layer['datasetId']] = get_images(self, dataset, layer, trainer)

                elif dataset['type'] == 'images_local':

                    all_memory = get_option(dataset['config'], 'allMemory', False, 'bool')

                    if all_memory:
                        datasets[layer['datasetId']] = read_images_in_memory(self, dataset, layer, trainer)
                    else:
                        datasets[layer['datasetId']] = read_images_keras_generator(self, dataset, layer, trainer)

                elif dataset['type'] == 'python':
                    name = dataset['id']

                    try:
                        sys.path.append(os.path.abspath('./aetros/dataset/' + os.path.dirname(name)))
                        data_provider = __import__(os.path.basename(name))
                        print("Imported dataset provider from %s " % (os.path.abspath('./aetros/dataset/' + name + '.py'),))

                        import inspect
                        argSpec = inspect.getargspec(data_provider.get_data)

                        if len(argSpec.args) > 0:
                            datasets[dataset['id']] = data_provider.get_data(trainer)
                        else:
                            datasets[dataset['id']] = data_provider.get_data()

                    except Exception:
                        trainer.logger.error('Could not import dataset code: ' + name + ' in ' + os.path.abspath('./aetros/dataset/'))
                        raise
                    finally:
                        sys.path.pop()

        return datasets

    def get_dataset_class_label(self, trainer, output_layer, prediction):
        dataset_id = output_layer['datasetId']
        dataset = self.datasets[dataset_id]

        if not dataset:
            raise Exception('Dataset of id %s does not exists. Available %s' % (
            dataset_id, ','.join(list(self.datasets.keys()))))

        if trainer.classes:
            return trainer.classes[prediction]

        elif dataset['type'] == 'python':
            name = dataset['id']
            datasets_dir = self.get_dataset_dir()
            try:
                sys.path.append(os.path.abspath('./aetros/dataset/' + os.path.dirname(name)))
                data_provider = __import__(os.path.basename(name))

                if hasattr(data_provider, 'get_class_label'):
                    return data_provider.get_class_label(prediction)
                else:
                    return prediction
            except Exception:
                print(("Method get_class_label failed in %s " % (datasets_dir + '/' + name + '.py',)))
                raise
            finally:
                sys.path.pop()


    def get_first_input_layer(self):
        config = self.job['config']
        return self.layers[0][0]

    def get_input_node(self, idx):
        config = self.job['config']
        return self.layers[0][idx]

    def get_first_output_layer(self):
        config = self.job['config']
        return self.layers[-1][0]

    def get_dataset(self, dataset_id):
        return self.datasets[dataset_id]

    def get_connected_layer(self, layers, to_net):
        connected_to_net = None
        for nets in layers:
            for net in nets:
                if 'connectedTo' in net:
                    for connectedTo in net['connectedTo']:
                        if connectedTo == to_net['id']:
                            return net

        return connected_to_net



import unittest
from aetros.utils import is_ignored, flatten_parameters, read_parameter_by_path


class TestAetrosDataset(unittest.TestCase):

    def testReadParameter(self):

        self.assertEqual(read_parameter_by_path({
            'lr': 0.5
        }, 'lr'), 0.5)

        self.assertEqual(read_parameter_by_path({
            'lr': "peter"
        }, 'lr'), "peter")

        self.assertEqual(read_parameter_by_path({
            'dataset': {
                'split': 0.1,
                'validation': 'jo'
            }
        }, 'dataset.split'), 0.1)

        self.assertEqual(read_parameter_by_path({
            'dataset': {
                'split': 0.1,
                'validation': 'jo'
            }
        }, 'dataset'), {
            'split': 0.1,
            'validation': 'jo'
        })

        choice_group = {
            'optimizer': {
                '$value': 'sgd',
                'sgd': {
                    'lr': 0.1
                }
            }
        }
        self.assertEqual(read_parameter_by_path(choice_group, 'optimizer'), 'sgd')
        self.assertEqual(read_parameter_by_path(choice_group, 'optimizer', return_group=True), {'lr': 0.1})
        self.assertEqual(read_parameter_by_path(choice_group, 'optimizer.lr'), 0.1)

    def testFlattenParameters(self):
        self.assertEqual(flatten_parameters({
            'lr': 0.5
        }), {
            'lr': 0.5
        })

        self.assertEqual(flatten_parameters({
            'list': ['peter', 'mayer']
        }), {
            'list': 'peter'
        })

        self.assertEqual(flatten_parameters({
            '_list': ['peter', 'mayer']
        }), {
            '_list': ['peter', 'mayer']
        })

        self.assertEqual(flatten_parameters({
            'optimizer': {
                'lr': 0.5
            }
        }), {
            'optimizer.lr': 0.5
        })

        self.assertEqual(flatten_parameters({
            'optimizer': {
                '$value': 'sgd',
                'sgd':  {
                    'lr': 0.5
                },
                'adadelta':  {
                    'lr': 0.5
                }
            }
        }), {
            'optimizer': "sgd",
            'optimizer.lr': 0.5,
        })

        self.assertEqual(flatten_parameters({
            'optimizer': {
                'sgd':  {
                    'lr': 0.5
                },
                'adadelta':  {
                    'lr': 0.6
                }
            }
        }), {
            'optimizer.sgd.lr': 0.5,
            'optimizer.adadelta.lr': 0.6,
        })

    def assertIgnored(self, path, patterns):
        self.assertTrue(is_ignored(path, patterns))

    def assertNotIgnored(self, path, patterns):
        self.assertFalse(is_ignored(path, patterns))

    def test_ignore(self):

        self.assertNotIgnored('script.py', '')
        self.assertNotIgnored('script.py', '!script.py')

        self.assertIgnored('script.py', 'script.py')
        self.assertIgnored('script.py', '*.py')
        self.assertNotIgnored('script.py', 'script.py\n!*.py')
        self.assertIgnored('script.py', 'script.py\n!*.py\npt.py')
        self.assertNotIgnored('script.py', 'script.py\n!*.py\n/pt.py')
        self.assertIgnored('/pt.py', 'script.py\n!*.py\n/pt.py')
        self.assertIgnored('script.py', '/*.py')

        self.assertIgnored('folder/script.py', 'script.py')
        self.assertIgnored('script.py', '/script.py')
        self.assertNotIgnored('folder/script.py', '/script.py')
        self.assertNotIgnored('pa-script.py', '/script.py')
        self.assertIgnored('pa-script.py', 'script.py')

        self.assertNotIgnored('script.py', 'folder/script.py')
        self.assertIgnored('folder/script.py', 'folder/script.py')
        self.assertIgnored('sub/folder/script.py', 'folder/script.py')

        self.assertIgnored('folder/script.py', '/folder/script.py')
        self.assertNotIgnored('sub/folder/script.py', '/folder/script.py')

        self.assertNotIgnored('folder/script.py', '/folder/script.py\n!script.py')
        self.assertIgnored('folder/script.py', '\n!*.py\n/folder/script.py')
        self.assertNotIgnored('folder/script.py', '/folder/script.py\n!*.py')
        self.assertIgnored('folder/script.py', '/folder/script.py\n!script.py\nscript.py')

        self.assertIgnored('folder/script.py', '/*/script.py')
        self.assertIgnored('folder/script.py', '/*er/script.py')

        self.assertNotIgnored('folder/script.py', '/*er2/script.py')
        self.assertIgnored('folder2/script.py', '/*er2/script.py')
        self.assertIgnored('folder/script.py', '/*er2/script.py\nfolder')

        self.assertIgnored('very/deep/datasets/dataset.zip', '*.zip')
        self.assertIgnored('very/deep/datasets/dataset.zip', 'datasets/*')
        self.assertNotIgnored('very/deep/datasets/dataset.zip', 'datasets/*\n!dataset.zip')

        self.assertNotIgnored('very/deep/datasets/dataset.zip', '/very/*/*.zip')
        self.assertIgnored('very/deep/datasets/dataset.zip', '/very/**/*.zip')
        self.assertNotIgnored('very/deep/dataset.zip', '/very/**/*.zip')
        self.assertIgnored('very/deep/dataset.zip', '/very/*/*.zip')
        self.assertIgnored('very/deep/dataset.zip', '/very/')





from __future__ import absolute_import
from __future__ import print_function
import argparse

import six
from aetros import api

class IdCommand:
    def __init__(self, logger):
        self.logger = logger
        self.client = None
        self.registered = False
        self.active = True

    def main(self, args):
        import aetros.const

        parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                                         prog=aetros.const.__prog__ + ' run')

        parsed_args = parser.parse_args(args)

        user = api.user()

        print("Key installed of account %s (%s)" % (user['username'], user['name']))

        if len(user['accounts']) > 0:
            for orga in six.itervalues(user['accounts']):
                print("  %s of organisation %s (%s)." % ("Owner" if orga['memberType'] == 1 else "Member", orga['username'], orga['name']))
        else:
            print("  Without membership to an organisation.")



from __future__ import absolute_import
from __future__ import print_function
import argparse
import subprocess

import sys

import os

from aetros.git import Git
from aetros.utils import read_config, unpack_simple_job_id

class AddCommand:
    def __init__(self, logger):
        self.logger = logger
        self.client = None
        self.registered = False
        self.active = True

    def main(self, args):
        import aetros.const

        parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter, prog=aetros.const.__prog__ + ' add')
        parser.add_argument('id', nargs='?', help="Model name like peter/mnist/ef8009d83a9892968097cec05b9467c685d45453")
        parser.add_argument('local_path', nargs='?', help="File path to be added to the job head commit history")
        parser.add_argument('git_path', nargs='?', help="Path under which the file will be stored.")
        parser.add_argument('-m', help="Commit message")

        parsed_args = parser.parse_args(args)

        if not parsed_args.id or not parsed_args.local_path:
            parser.print_help()
            sys.exit(1)

        config = read_config()

        [model, job_id] = unpack_simple_job_id(parsed_args.id)
        ref = 'refs/aetros/job/' + job_id

        git_dir = os.path.normpath(config['storage_dir'] + '/' + model + '.git')

        if not os.path.isdir(git_dir):
            self.logger.error("Git repository for model %s in %s not found." % (parsed_args.id, git_dir))
            self.logger.error("Are you in the correct directory?")

        git = Git(self.logger, None, config, model)
        git.job_id = job_id
        try:
            git.read_tree(ref)

            with open(parsed_args.local_path, 'rb') as f:
                id = git.commit_file(parsed_args.m or 'Added file', parsed_args.git_path or parsed_args.local_path, f.read())

                self.logger.info('Successfully committed to ' + id + '.')
                self.logger.info('Run "aetros push-job ' + parsed_args.id + '" to upload changes to AETROS.')
        finally:
            git.clean_up()


"""
Copyright (c) 2001, 2002 Enthought, Inc.
All rights reserved.

Copyright (c) 2003-2017 SciPy Developers.
All rights reserved.

A collection of image utilities using the Python Imaging Library (PIL).

Note that PIL is not a dependency of SciPy and this module is not
available on systems that don't have PIL installed.

aetros-cli: This source is coming from the scipy package and is here copied since building the
scipy package on various computers takes a long time and is not necessary to have
such a big package as requirement when you only need 2-3 methods. Also scipy has a lot of different
licenes in their package, so we used here only code with the MIT license.
"""


import numpy

from numpy import (amin, amax, ravel, asarray, cast, arange, ones, newaxis,
                   transpose, iscomplexobj, uint8, issubdtype, array)

try:
    from PIL import Image, ImageFilter
except ImportError:
    import Image
    import ImageFilter


def fromimage(im, flatten=False, mode=None):
    """
    Return a copy of a PIL image as a numpy array.

    Parameters
    ----------
    im : PIL image
        Input image.
    flatten : bool
        If true, convert the output to grey-scale.
    mode : str, optional
        Mode to convert image to, e.g. ``'RGB'``.  See the Notes of the
        `imread` docstring for more details.

    Returns
    -------
    fromimage : ndarray
        The different colour bands/channels are stored in the
        third dimension, such that a grey-image is MxN, an
        RGB-image MxNx3 and an RGBA-image MxNx4.

    """
    if not Image.isImageType(im):
        raise TypeError("Input is not a PIL image.")

    if mode is not None:
        if mode != im.mode:
            im = im.convert(mode)
    elif im.mode == 'P':
        # Mode 'P' means there is an indexed "palette".  If we leave the mode
        # as 'P', then when we do `a = array(im)` below, `a` will be a 2-D
        # containing the indices into the palette, and not a 3-D array
        # containing the RGB or RGBA values.
        if 'transparency' in im.info:
            im = im.convert('RGBA')
        else:
            im = im.convert('RGB')

    if flatten:
        im = im.convert('F')
    elif im.mode == '1':
        # Workaround for crash in PIL. When im is 1-bit, the call array(im)
        # can cause a seg. fault, or generate garbage. See
        # https://github.com/scipy/scipy/issues/2138 and
        # https://github.com/python-pillow/Pillow/issues/350.
        #
        # This converts im from a 1-bit image to an 8-bit image.
        im = im.convert('L')

    a = array(im)
    return a

_errstr = "Mode is unknown or incompatible with input array shape."


# Returns a byte-scaled image
def bytescale(data, cmin=None, cmax=None, high=255, low=0):
    """
    Byte scales an array (image).

    Byte scaling means converting the input image to uint8 dtype and scaling
    the range to ``(low, high)`` (default 0-255).
    If the input image already has dtype uint8, no scaling is done.

    Parameters
    ----------
    data : ndarray
        PIL image data array.
    cmin : scalar, optional
        Bias scaling of small values. Default is ``data.min()``.
    cmax : scalar, optional
        Bias scaling of large values. Default is ``data.max()``.
    high : scalar, optional
        Scale max value to `high`.  Default is 255.
    low : scalar, optional
        Scale min value to `low`.  Default is 0.

    Returns
    -------
    img_array : uint8 ndarray
        The byte-scaled array.

    Examples
    --------
    >>> from scipy.misc import bytescale
    >>> img = np.array([[ 91.06794177,   3.39058326,  84.4221549 ],
    ...                 [ 73.88003259,  80.91433048,   4.88878881],
    ...                 [ 51.53875334,  34.45808177,  27.5873488 ]])
    >>> bytescale(img)
    array([[255,   0, 236],
           [205, 225,   4],
           [140,  90,  70]], dtype=uint8)
    >>> bytescale(img, high=200, low=100)
    array([[200, 100, 192],
           [180, 188, 102],
           [155, 135, 128]], dtype=uint8)
    >>> bytescale(img, cmin=0, cmax=255)
    array([[91,  3, 84],
           [74, 81,  5],
           [52, 34, 28]], dtype=uint8)

    """
    if data.dtype == uint8:
        return data

    if high > 255:
        raise ValueError("`high` should be less than or equal to 255.")
    if low < 0:
        raise ValueError("`low` should be greater than or equal to 0.")
    if high < low:
        raise ValueError("`high` should be greater than or equal to `low`.")

    if cmin is None:
        cmin = data.min()
    if cmax is None:
        cmax = data.max()

    cscale = cmax - cmin
    if cscale < 0:
        raise ValueError("`cmax` should be larger than `cmin`.")
    elif cscale == 0:
        cscale = 1

    scale = float(high - low) / cscale
    bytedata = (data - cmin) * scale + low
    return (bytedata.clip(low, high) + 0.5).astype(uint8)


def toimage(arr, high=255, low=0, cmin=None, cmax=None, pal=None,
            mode=None, channel_axis=None):
    """Takes a numpy array and returns a PIL image.

    The mode of the PIL image depends on the array shape and the `pal` and
    `mode` keywords.

    For 2-D arrays, if `pal` is a valid (N,3) byte-array giving the RGB values
    (from 0 to 255) then ``mode='P'``, otherwise ``mode='L'``, unless mode
    is given as 'F' or 'I' in which case a float and/or integer array is made.

    Notes
    -----
    For 3-D arrays, the `channel_axis` argument tells which dimension of the
    array holds the channel data.

    For 3-D arrays if one of the dimensions is 3, the mode is 'RGB'
    by default or 'YCbCr' if selected.

    The numpy array must be either 2 dimensional or 3 dimensional.

    """
    data = asarray(arr)
    if iscomplexobj(data):
        raise ValueError("Cannot convert a complex-valued array.")
    shape = list(data.shape)
    valid = len(shape) == 2 or ((len(shape) == 3) and
                                ((3 in shape) or (4 in shape)))
    if not valid:
        raise ValueError("'arr' does not have a suitable array shape for "
                         "any mode.")
    if len(shape) == 2:
        shape = (shape[1], shape[0])  # columns show up first
        if mode == 'F':
            data32 = data.astype(numpy.float32)
            image = Image.frombytes(mode, shape, data32.tostring())
            return image
        if mode in [None, 'L', 'P']:
            bytedata = bytescale(data, high=high, low=low,
                                 cmin=cmin, cmax=cmax)
            image = Image.frombytes('L', shape, bytedata.tostring())
            if pal is not None:
                image.putpalette(asarray(pal, dtype=uint8).tostring())
                # Becomes a mode='P' automagically.
            elif mode == 'P':  # default gray-scale
                pal = (arange(0, 256, 1, dtype=uint8)[:, newaxis] *
                       ones((3,), dtype=uint8)[newaxis, :])
                image.putpalette(asarray(pal, dtype=uint8).tostring())
            return image
        if mode == '1':  # high input gives threshold for 1
            bytedata = (data > high)
            image = Image.frombytes('1', shape, bytedata.tostring())
            return image
        if cmin is None:
            cmin = amin(ravel(data))
        if cmax is None:
            cmax = amax(ravel(data))
        data = (data*1.0 - cmin)*(high - low)/(cmax - cmin) + low
        if mode == 'I':
            data32 = data.astype(numpy.uint32)
            image = Image.frombytes(mode, shape, data32.tostring())
        else:
            raise ValueError(_errstr)
        return image

    # if here then 3-d array with a 3 or a 4 in the shape length.
    # Check for 3 in datacube shape --- 'RGB' or 'YCbCr'
    if channel_axis is None:
        if (3 in shape):
            ca = numpy.flatnonzero(asarray(shape) == 3)[0]
        else:
            ca = numpy.flatnonzero(asarray(shape) == 4)
            if len(ca):
                ca = ca[0]
            else:
                raise ValueError("Could not find channel dimension.")
    else:
        ca = channel_axis

    numch = shape[ca]
    if numch not in [3, 4]:
        raise ValueError("Channel axis dimension is not valid.")

    bytedata = bytescale(data, high=high, low=low, cmin=cmin, cmax=cmax)
    if ca == 2:
        strdata = bytedata.tostring()
        shape = (shape[1], shape[0])
    elif ca == 1:
        strdata = transpose(bytedata, (0, 2, 1)).tostring()
        shape = (shape[2], shape[0])
    elif ca == 0:
        strdata = transpose(bytedata, (1, 2, 0)).tostring()
        shape = (shape[2], shape[1])
    if mode is None:
        if numch == 3:
            mode = 'RGB'
        else:
            mode = 'RGBA'

    if mode not in ['RGB', 'RGBA', 'YCbCr', 'CMYK']:
        raise ValueError(_errstr)

    if mode in ['RGB', 'YCbCr']:
        if numch != 3:
            raise ValueError("Invalid array shape for mode.")
    if mode in ['RGBA', 'CMYK']:
        if numch != 4:
            raise ValueError("Invalid array shape for mode.")

    # Here we know data and mode is correct
    image = Image.frombytes(mode, shape, strdata)
    return image



def imresize(arr, size, interp='bilinear', mode=None):
    """
    Resize an image.

    Parameters
    ----------
    arr : ndarray
        The array of image to be resized.

    size : int, float or tuple
        * int   - Percentage of current size.
        * float - Fraction of current size.
        * tuple - Size of the output image.

    interp : str, optional
        Interpolation to use for re-sizing ('nearest', 'lanczos', 'bilinear', 'bicubic'
        or 'cubic').

    mode : str, optional
        The PIL image mode ('P', 'L', etc.) to convert `arr` before resizing.

    Returns
    -------
    imresize : ndarray
        The resized array of image.

    See Also
    --------
    toimage : Implicitly used to convert `arr` according to `mode`.
    scipy.ndimage.zoom : More generic implementation that does not use PIL.

    """
    im = toimage(arr, mode=mode)
    ts = type(size)
    if issubdtype(ts, int):
        percent = size / 100.0
        size = tuple((array(im.size)*percent).astype(int))
    elif issubdtype(type(size), float):
        size = tuple((array(im.size)*size).astype(int))
    else:
        size = (size[1], size[0])
    func = {'nearest': 0, 'lanczos': 1, 'bilinear': 2, 'bicubic': 3, 'cubic': 3}
    imnew = im.resize(size, resample=func[interp])
    return fromimage(imnew)


from __future__ import absolute_import

import sys
import six
from threading import Timer, Thread, Lock


def drain_stream(stream, decode='utf-8'):
    content = six.b('')

    while True:
        try:
            # read() needs to block
            # buf = os.read(buffer.fileno(), 4096)
            buf = stream.read(1)
            if buf == six.b(''):
                break
            content += buf
        except Exception:
            break

    if decode:
        return content.decode(decode)

    return content


class GeneralLogger(object):
    def __init__(self, redirect_to, job_backend=None):
        self.job_backend = job_backend
        self.buffer = ''
        self.last_timer = None
        self.last_messages = ''
        self.logger = redirect_to or sys.__stdout__
        self.lock = Lock()
        self.attach_last_messages = {}
        self.buffer_disabled = False

    def disable_buffer(self):
        self.buffer_disabled = True
        self.buffer = ''

    def clear_buffer(self):
        self.buffer = ''

    def fileno(self):
        return self.logger.fileno()

    def isatty(self):
        return self.logger.isatty()

    def flush(self):
        self.logger.flush()
        self.send_buffer()

    def send_buffer(self):
        self.last_timer = None

        if self.buffer:
            if self.job_backend:
                if self.job_backend.write_log(self.buffer):
                    self.buffer = ''

    def attach(self, buffer, read_line=False):
        """
        Read buffer until end (read() returns '') and sends it to self.logger and self.job_backend.

        :param buffer: a buffer instance with block read() method
        """

        bid = id(buffer)
        self.attach_last_messages[bid] = six.b('')

        lock = Lock()

        def reader():
            lock.acquire()
            while True:

                try:
                    # read() needs to block
                    # buf = os.read(buffer.fileno(), 4096)
                    if read_line:
                        buf = buffer.readline()
                    else:
                        buf = buffer.read(1)
                    if buf == six.b(''):
                        break

                    self.attach_last_messages[bid] += buf

                    if len(self.attach_last_messages[bid]) > 21 * 1024:
                        self.attach_last_messages[bid] = self.attach_last_messages[bid][-20 * 1024:]

                    self.write(buf)
                except ValueError as e:
                    if 'operation on closed' in e.message:
                        break

                except Exception as e:
                    # we need to make sure, we continue to read otherwise the process of this buffer
                    # will block and we have a stuck process.
                    sys.__stderr__.write(str(type(e))+': ' + str(e))
                    pass

            lock.release()

        thread = Thread(target=reader)
        thread.daemon = True
        thread.start()

        def wait():
            lock.acquire()
            self.send_buffer()
            lock.release()

        return wait

    def write(self, message):
        try:
            self.lock.acquire()

            if hasattr(message, 'decode'):
                # don't decode string again
                # necessary for Python3
                message = message.decode('utf-8')

            self.logger.write(message)
            self.logger.flush()

            self.last_messages += message
            if len(self.last_messages) > 20 * 1024:
                self.last_messages = self.last_messages[-20 * 1024:]

            if not self.buffer_disabled:
                for char in message:
                    if '\b' == char and self.buffer:
                        self.buffer = self.buffer[:-1]
                    else:
                        self.buffer += char

                if not self.last_timer:
                    self.last_timer = Timer(1.0, self.send_buffer)
                    self.last_timer.start()
        except Exception as e:
            sys.__stderr__.write(str(e))
        finally:
            self.lock.release()



import subprocess

import re

import os


def silent_execute(args):
    try:
        with open(os.devnull, 'r+b', 0) as DEVNULL:
            return subprocess.check_output(args, stderr=DEVNULL).decode("utf-8").strip()
    except Exception:
        return None


def get_branches():
    output = silent_execute(['git', 'branch'])
    if output:
        branches = output.split('\n')
    
        return [x.strip(' *') for x in branches]


def get_current_branch():
    output = silent_execute(['git', 'branch'])
    if output:
        m = re.search('\* ([^\s]+)', output)
        current_branch = m.group(1) if m else None

        return current_branch


def get_current_remote_url(origin_name = 'origin'):
    output = silent_execute(['git', 'remote', '-v'])
    if output:
        import re
        regex = re.compile('^' + re.escape(origin_name) + '\t([^\s]+)', re.MULTILINE)
        matches = [m.groups() for m in regex.finditer(output)]
        if matches:
            return matches[0][0]

        return None


def get_current_commit_hash():
    return silent_execute(['git', 'rev-parse', 'HEAD'])


def get_current_commit_message():
    return silent_execute(['git', 'log', '-1', '--pretty=%B', 'HEAD'])


def get_current_commit_author():
    return silent_execute(['git', 'log', '-1', '--pretty=%an <%ae>', 'HEAD'])



from __future__ import division
from __future__ import absolute_import

import os
import re
import time
import datetime

import numpy as np
import signal

import paramiko
import six
import sys
import ruamel.yaml as yaml

start_time = time.time()
last_time = None

def stop_time(title=''):
    global last_time
    diff = ("{0:.10f}".format(time.time() - last_time)) if last_time is not None else ''
    last_time = time.time()
    sys.__stdout__.write("STOP_TIME: " + str(time.time()-start_time) + "s - diff: "+diff+"  - " +title+ "\n")


def get_option(dict, key, default=None, type=None):
    if key not in dict or dict[key] == '':
        return default

    if type == 'bool':
        bool(dict[key])

    return dict[key]


def unpack_simple_job_id(full_id):
    [owner, model, id] = unpack_full_job_id(full_id)

    return [owner + '/' + model, id]


def unpack_full_job_id(full_id):
    if full_id.count('/') < 2:
        raise Exception('Not a valid full job id: ' + full_id)

    owner = full_id[0:full_id.index('/')]
    model = full_id[full_id.index('/') + 1:full_id.rindex('/')]
    id = full_id[full_id.rindex('/')+1:]

    return [owner, model, id]


def create_ssh_stream(config, exit_on_failure=True):
    ssh_stream = paramiko.client.SSHClient()
    # ssh_stream.load_system_host_keys()
    ssh_stream.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    key = None
    key_filename = get_ssh_key_for_host(config['host'])

    if config['ssh_key_base64']:
        key_filename = None
        key = paramiko.RSAKey.from_private_key(six.StringIO(config['ssh_key_base64']))

    if not key and not key_filename:
        raise Exception("No SSH key configured for " + config['host'] + ". See https://aetros.com/docu/trainer/authentication")

    key_description = key_filename if key_filename else 'from server'

    try:
        ssh_stream.connect(config['host'], key_filename=key_filename, username='git', compress=True, pkey=key)
    except Exception as e:
        if isinstance(e, paramiko.ssh_exception.AuthenticationException) or isinstance(e, paramiko.ssh_exception.SSHException):
            if exit_on_failure:
                print("Fatal: AETROS authentication against "+config['host']+" failed using key "+key_description+": "+str(e)+
                      ". Did you setup SSH keys correctly? See https://aetros.com/docu/trainer/authentication")
                sys.exit(1)
        raise

    return ssh_stream


def setup_git_ssh(config):
    import tempfile
    ssh_command = config['ssh']
    ssh_command += ' -o StrictHostKeyChecking=no'

    ssh_key = None
    if config['ssh_key_base64']:
        ssh_key = tempfile.NamedTemporaryFile(delete=False, prefix='ssh_key_')
        ssh_key.write(six.b(config['ssh_key_base64']))
        ssh_key.close()
        ssh_command += ' -i '+ ssh_key.name
        os.chmod(ssh_key.name, 0o600)
    # elif config['ssh_key']:
    #     ssh_command += ' -i '+ os.path.expanduser(config['ssh_key'])

    ssh_script = tempfile.NamedTemporaryFile(delete=False, prefix='git_ssh_')
    ssh_script.write(six.b(ssh_command + ' "$@"'))
    ssh_script.close()
    os.environ['GIT_SSH'] = ssh_script.name
    os.chmod(ssh_script.name, 0o700)

    def delete():
        if os.path.exists(ssh_script.name):
            os.unlink(ssh_script.name)

        if ssh_key and os.path.exists(ssh_key.name):
            os.unlink(ssh_key.name)

    return delete


def get_ssh_key_for_host(host):
    ssh_config = paramiko.SSHConfig()
    user_config_file = os.path.expanduser("~/.ssh/config")
    if os.path.exists(user_config_file):
        with open(user_config_file) as f:
            ssh_config.parse(f)

    user_config = ssh_config.lookup(host)

    if 'identityfile' in user_config:
        path = os.path.expanduser(user_config['identityfile'][0])
        if not os.path.exists(path):
            raise Exception("Specified IdentityFile "+path
                            + " for " + host + " in ~/.ssh/config not existing anymore.")

        return path


def read_home_config(path = None, logger=None):
    if not path:
        path = os.getenv('AETROS_HOME_CONFIG_FILE') or '~/aetros.yml'

    path = os.path.normpath(os.path.expanduser(path))
    custom_config = {}

    if os.path.exists(path):
        f = open(path, 'r')
        try:
            logger and logger.debug('Home config loaded from ' + os.path.realpath(path))
            custom_config = yaml.safe_load(f)
        except Exception:
            raise Exception('Could not load aetros home config at ' + os.path.realpath(path))

        if custom_config is None:
            custom_config = {}

    config = {
        'host': os.getenv('AETROS_HOST') or 'trainer.aetros.com',
        'storage_dir': os.getenv('AETROS_STORAGE_DIR') or '~/.aetros',
        'ssh_key_base64': os.getenv('AETROS_SSH_KEY_BASE64'),
        'image': None,
        'ssh': 'ssh',
        'git': 'git',
        'docker': 'docker',
        'docker_options': [],
        'ssl_verify': True,
    }

    config.update(custom_config)

    config['storage_dir'] = os.path.abspath(os.path.expanduser(config['storage_dir']))

    return config

ignore_pattern_cache = {}


def extract_parameters(full_definitions, overwritten = None, incoming_path = ''):
    container = {}

    if overwritten is None:
        overwritten = {}

    for parameter in full_definitions:
        param_type = parameter['type']

        if 'defaultValue' in parameter:
            defaultValue = parameter['defaultValue']
        elif 'default' in parameter:
            defaultValue = parameter['default']
        else:
            defaultValue = None

        value = defaultValue
        if 'value' in parameter:
            value = parameter['value']

        name = parameter['name']

        path = (incoming_path + '.' + name).strip('.')

        if path in overwritten:
            value = overwritten[path]

        if param_type == 'string':
            container[name] = str(value)
#
        if param_type == 'number':
            container[name] = float(value)
#
        if param_type == 'boolean':
            container[name] = bool(value)
#
        if param_type == 'array':
            container[name] = value

        if param_type == 'group':
            container[name] = extract_parameters(parameter['children'], overwritten, path)

        if param_type == 'choice_group':
            if '' == value:
                #no default value nor value set
                continue

            if isinstance(value, six.string_types) and value not in parameter['children']:
                found_names = []
                for idx, child in enumerate(parameter['children']):
                    found_names.append(child['name'])
                    if value == child['name']:
                        value = idx
                        break

                if isinstance(value, six.string_types):
                    names = ', '.join(found_names)
                    raise Exception(str(value) + " is not available in " + path + ". " + names)

            value_name = parameter['children'][value]['name']
            container[name] = {
                '$value': value_name,
            }
            container[name][value_name] = extract_parameters(parameter['children'][value]['children'], overwritten, path)

        if param_type == 'choice_string' or param_type == 'choice_number':
            if path in overwritten:
                for idx, child in enumerate(parameter['children']):
                    if overwritten[path] == child['value']:
                        value = idx
                        break

            if value is None or (value < 0 or value >= len(parameter['children'])):
                value = 0

            container[name] = parameter['children'][value]['value']

    return container


def lose_parameters_to_full(parameters):
    if not parameters:
        return {}

    if isinstance(parameters, list):
        # full definition with type, name etc.
        return parameters

    def type_of(v):
        if isinstance(v, six.string_types):
            return 'string'

        if isinstance(v, (float, int)):
            return 'number'

        if isinstance(v, bool):
            return 'boolean'

        if isinstance(v, (list, dict)):
            return 'array'

    def extract_from_parameters(parameters):
        definitions = []

        for k, v in six.iteritems(parameters):
            definition = {'name': k}

            if k[0] != '_' and isinstance(v, list):
                # //regular array ['a', 'b']
                if not v:
                    definition['type'] = 'choice_string'
                    continue
                if isinstance(v[0], six.string_types):
                    definition['type'] = 'choice_string'
                else:
                    definition['type'] = 'choice_number'

                definition['children'] = []
                definition['defaultValue'] = 0

                for child in v:
                    definition['children'].append({'value': child})

            elif k[0] != '_' and isinstance(v, dict):
                # named keys, so a group
                # ['a' => 'bla']
                # becomes a choicegroup when all childern are groups as well
                all_groups = len(v) > 1

                for children in six.itervalues(v):
                    if not isinstance(children, dict):
                        all_groups = False
                        break

                if all_groups:
                    # when group in group, then first become choice_group
                    definition['type'] = 'choice_group'
                    definition['defaultValue'] = 0
                else:
                    definition['type'] = 'group'

                definition['children'] = extract_from_parameters(v)
            else:
                definition['type'] = type_of(v)
                definition['defaultValue'] = v

            definitions.append(definition)

        return definitions

    return extract_from_parameters(parameters)


def read_parameter_by_path(dictionary, path, return_group=False):

    # if path in dictionary:
    #     return dictionary[path]

    # elif '.' not in path:
    #     raise Exception('Parameter ' + path + ' not found and no default value given.')

    if not dictionary:
        return None

    current_group = None

    current = dictionary

    for item in path.split('.'):
        current_group = None
        if item not in current:
            raise Exception('Parameter ' + str(path) + ' not found and no default value given.')

        current = current[item]

        if isinstance(current, dict) and '$value' in current and current['$value'] in current:
            current_group = current
            current = current[current['$value']]

    if isinstance(current_group, dict) and '$value' in current_group and current_group['$value'] in current_group:
        if return_group:
            return current_group[current_group['$value']]
        else:
            return current_group['$value']

    return current


def flatten_parameters(parameters, incomingPath = ''):
    result = {}

    for key, value in six.iteritems(parameters):
        if '$value' == key: continue

        path = (incomingPath + '.' + key).strip('.')

        if isinstance(value, dict) and key[0] != '_':
            if '$value' in value:
                result[path] = value['$value']
                if value['$value'] not in value:
                    raise Exception(value['$value'] + " of " + path + " not in " + str(value))
                value = value[value['$value']]
                result.update(flatten_parameters(value, path))
            else:
                result.update(flatten_parameters(value, path))

        elif isinstance(value, list) and key[0] != '_':
            result[path] = value[0]
        else:
            result[path] = value

    return result


def is_ignored(path, ignore_patters):
    if isinstance(ignore_patters, six.string_types):
        ignore_patters = ignore_patters.split('\n')

    ignored = None

    if not ignore_patters:
        return False

    for pattern in ignore_patters:
        if not pattern:
            continue

        if pattern in ignore_pattern_cache:
            reobj = ignore_pattern_cache[pattern]
        else:
            if pattern[0] == '!':
                regex = re.escape(pattern[1:])
            else:
                regex = re.escape(pattern)

            regex = regex.replace('\\*\\*', '([^/\\\\]+[//\\\\])+([^/\\\\]+)')
            regex = regex.replace('\\*', '[^/\\\\]+')
            regex = '(' + regex + ')'
            if pattern[0] == '/' or (pattern[0] == '!' and pattern[1] == '/'):
                regex = '^' + regex
            else:
                regex = '^.*' + regex

            reobj = re.compile(regex)
            ignore_pattern_cache[pattern] = reobj

        normalized_path = path

        if pattern.startswith('!'):
            if pattern[1] == '/' and path[0] != '/':
                normalized_path = '/' + path

            # whitelist begins with !
            if reobj.match(normalized_path):
                ignored = False
        else:
            if pattern[0] == '/' and path[0] != '/':
                normalized_path = '/' + path

            if reobj.match(normalized_path):
                ignored = True

    if ignored is None:
        return False

    return ignored


def read_config(path = 'aetros.yml', logger=None):
    path = os.path.normpath(os.path.expanduser(path))

    config = {
        'dockerfile': None,
        'command': None,
        'install': None,
        'ignore': None,
        'image': None,
        'server': None,
        'parameters': {},
        'servers': None,
        'before_command': [],
    }

    if os.path.exists(path):
        f = open(path, 'r')

        custom_config = yaml.safe_load(f)
        if custom_config is None:
            custom_config = {}

        if 'storage_dir' in custom_config:
            del custom_config['storage_dir']

        config.update(custom_config)

        logger and logger.debug('Config loaded from ' + os.path.realpath(path))

    if 'parameters' not in config:
        config['parameters'] = {}

    return config


def raise_sigint():
    """
    Raising the SIGINT signal in the current process and all sub-processes.

    os.kill() only issues a signal in the current process (without subprocesses).
    CTRL+C on the console sends the signal to the process group (which we need).
    """
    if hasattr(signal, 'CTRL_C_EVENT'):
        # windows. Need CTRL_C_EVENT to raise the signal in the whole process group
        os.kill(os.getpid(), signal.CTRL_C_EVENT)
    else:
        # unix.
        pgid = os.getpgid(os.getpid())
        if pgid == 1:
            os.kill(os.getpid(), signal.SIGINT)
        else:
            os.killpg(os.getpgid(os.getpid()), signal.SIGINT)


def prepend_signal_handler(sig, f):

    previous_handler = None
    if callable(signal.getsignal(sig)):
        previous_handler = signal.getsignal(sig)

    def execute_signal_handler(*args, **kwargs):
        f(*args, **kwargs)

        # previous handles comes after this, so we actually prepend
        if previous_handler:
            previous_handler(*args, **kwargs)

    signal.signal(sig, execute_signal_handler)


def invalid_json_values(obj):
    if isinstance(obj, np.generic):
        return obj.item()
    if isinstance(obj, np.ndarray):
        return obj.tolist()
    if isinstance(obj, bytes):
        return obj.decode('cp437')

    if isinstance(map, type) and isinstance(obj, map):
        # python 3 map
        return list(obj)

    raise TypeError('Invalid data type passed to json encoder: ' + type(obj).__name__)


def get_time(self):
    ts = time.time()
    st = datetime.datetime.fromtimestamp(ts).strftime('%Y-%m-%d %H:%M:%S')

    return st


def get_time_with_milli(self):
    ts = time.time()
    st = datetime.datetime.fromtimestamp(ts).strftime('%Y-%m-%d %H:%M:%S')

    return st + '.' + str(ts % 1)[2:6]


def human_size(size_bytes, precision=0):
    """
    Format a size in bytes into a 'human' file size, e.g. bytes, KB, MB, GB, TB, PB
    Note that bytes/KB will be reported in whole numbers but MB and above will have greater precision
    e.g. 1 byte, 43 bytes, 443 KB, 4.3 MB, 4.43 GB, etc
    """
    if size_bytes == 1:
        # because I really hate unnecessary plurals
        return "1 byte"

    suffixes_table = [('bytes',0),('KB',0),('MB',1),('GB',2),('TB',2), ('PB',2)]

    num = float(size_bytes)
    for suffix, precision in suffixes_table:
        if num < 1024.0:
            break
        num /= 1024.0

    if precision == 0:
        formatted_size = "%d" % num
    else:
        formatted_size = str(round(num, ndigits=precision))

    return "%s %s" % (formatted_size, suffix)


def array_to_img(x, scale=True):
    """
    x should be shape (channels, width, height)
    """
    from PIL import Image
    if x.ndim != 3:
        raise Exception('Unsupported shape : ', str(x.shape), '. Need (channels, width, height)')
    if scale:
        x += max(-np.min(x), 0)
        x /= np.max(x)
        x *= 255
    if x.shape[0] == 3:
        # RGB
        if x.dtype != 'uint8':
            x = x.astype('uint8')
        return Image.fromarray(x.astype('uint8'), 'RGB')
    elif x.shape[0] == 1:
        # grayscale
        if x.dtype != 'uint8':
            x = x.astype('uint8')
        return Image.fromarray(x.reshape(x.shape[1], x.shape[2]), 'L')
    else:
        raise Exception('Unsupported channel number: ', x.shape[0])



from __future__ import print_function
from __future__ import absolute_import

import json
import numpy as np
import os

from aetros.utils import unpack_full_job_id
from .backend import JobBackend, invalid_json_values

def predict(logger, job_id, file_paths, weights_path=None):

    owner, name, id = unpack_full_job_id(job_id)

    job_backend = JobBackend(model_name=owner+'/'+name)
    job_backend.fetch(id)
    job_backend.load(id)

    job_model = job_backend.get_job_model()
    os.chdir(job_backend.git.work_tree)

    if not weights_path:
        weights_path = job_model.get_weights_filepath_latest()

    from .Trainer import Trainer

    trainer = Trainer(job_backend)
    job_model.set_input_shape(trainer)

    import keras.backend
    if hasattr(keras.backend, 'set_image_dim_ordering'):
        keras.backend.set_image_dim_ordering('tf')

    if hasattr(keras.backend, 'set_image_data_format'):
        keras.backend.set_image_data_format('channels_last')

    job_backend.logger.info("Load model and compile ...")

    model = job_model.get_built_model(trainer)
    trainer.model = model

    from aetros.keras import load_weights
    logger.info('Load weights from ' + weights_path)
    load_weights(model, weights_path)

    inputs = []
    for idx, file_path in enumerate(file_paths):
        inputs.append(job_model.convert_file_to_input_node(file_path, job_model.get_input_node(idx)))

    job_backend.logger.info("Start prediction ...")

    prediction = job_model.predict(trainer, np.array(inputs))

    print(json.dumps(prediction, indent=4, default=invalid_json_values))



from __future__ import absolute_import
from __future__ import print_function
import argparse

import sys

from aetros.starter import start_keras
from aetros.backend import JobBackend
from aetros.utils import unpack_full_job_id


class StartSimpleCommand:
    def __init__(self, logger):
        self.logger = logger
        self.client = None
        self.registered = False
        self.active = True

    def main(self, args):
        import aetros.const

        parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                                         prog=aetros.const.__prog__ + ' run', description="Internal usage.")

        parser.add_argument('id', nargs='?', help='Job id, e.g. user/modelname/0db75a64acb74c27bd72c22e359de7a4c44a20e5 to start a pre-created job.')
        parser.add_argument('--fetch', action='store_true', help="Fetch job from server.")

        parsed_args = parser.parse_args(args)

        if not parsed_args.id:
            parser.print_help()
            sys.exit(1)

        owner, name, id = unpack_full_job_id(parsed_args.id)

        job_backend = JobBackend(model_name=owner + '/' + name)
        job_backend.section('checkout')

        if parsed_args.fetch:
            job_backend.fetch(id)

        job_backend.load(id)
        job_backend.start()

        start_keras(self.logger, job_backend)



from __future__ import absolute_import
from __future__ import print_function
import argparse

import sys

import os

import aetros.utils.git
from aetros.logger import GeneralLogger

from aetros.starter import start

from aetros.backend import JobBackend
from aetros import api
from aetros.utils import read_config, human_size, lose_parameters_to_full, extract_parameters, stop_time


class RunCommand:
    def __init__(self, logger):
        self.logger = logger
        self.client = None
        self.registered = False
        self.active = True

    def main(self, args):
        import aetros.const

        parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                                         prog=aetros.const.__prog__ + ' run')
        parser.add_argument('command', nargs='?', help="The command to run. Default read in aetros.yml")
        parser.add_argument('-i', '--image', help="Which Docker image to use for the command. Default read in aetros.yml. If not specified, command is executed on the host.")
        parser.add_argument('-s', '--server', action='append', help="Limits the server pool to this server. Default not limitation or read in aetros.yml. Multiple --server allowed.")
        parser.add_argument('-m', '--model', help="Under which model this job should be listed. Default read in aetros.yml")
        parser.add_argument('-l', '--local', action='store_true', help="Start the job immediately on the current machine.")
        parser.add_argument('-c', '--config', help="Default aetros.yml in current working directory.")
        parser.add_argument('--priority', help="Increases or decreases priority. Default is 0.")

        parser.add_argument('--cpu', help="How many CPU cores should be assigned to job. Docker only.")
        parser.add_argument('--memory', help="How much memory should be assigned to job. Docker only.")
        parser.add_argument('--gpu', help="How many GPU cards should be assigned to job. Docker only.")
        parser.add_argument('--gpu_memory', help="Memory requirement for the GPU. Docker only.")

        parser.add_argument('--max-time', help="Limit execution time in seconds. Sends SIGINT to the process group when reached.")
        parser.add_argument('--max-epochs', help="Limit execution epochs. Sends SIGINT to the process group when reached.")

        parser.add_argument('--gpu-device', action='append', help="Which device id should be mapped into the NVIDIA docker container.")

        parser.add_argument('--volume', '-v', action='append', help="Volume into docker")
        parser.add_argument('-e', action='append', help="Sets additional environment variables. '-e name=value' to set value, or '-e name' to read from current env")

        parser.add_argument('-p', '--param', action='append', help="Sets a hyperparameter, example '--param name=value'. Multiple --param allowed.")

        parsed_args = parser.parse_args(args)

        config = read_config(parsed_args.config or 'aetros.yml')

        env = {}
        if parsed_args.e:
            for item in parsed_args.e:
                if '=' in item:
                    k, v = item.split('=')
                else:
                    k = item
                    v = os.getenv(k)
                env[k] = v

        if 'command' not in config and not parsed_args.command:
            self.logger.error('No "command" given in aetros.yml or as argument.')
            sys.exit(1)

        job = JobBackend(parsed_args.model, self.logger, parsed_args.config or 'aetros.yml')
        ignore = []
        if 'ignore' in config:
            ignore = config['ignore']
        job.job = {'config': {'ignore': ignore}}

        files_added, size_added = job.add_files()

        print("%d files added (%s)" % (files_added, human_size(size_added, 2)))

        create_info = {
            'type': 'custom',
            'config': config
        }

        incoming_hyperparameter = {}
        if parsed_args.param:
            for param in parsed_args.param:
                if '=' not in param:
                    raise Exception('--param ' + param + ' does not contain a `=`. Please use "--param name=value"')

                name, value = param.split('=')
                incoming_hyperparameter[name] = value

        # first transform simple format in the full definition with parameter types
        # (string, number, group, choice_group, etc)
        full_hyperparameters = lose_parameters_to_full(config['parameters'])

        # now extract hyperparameters from full definition, and overwrite stuff using
        # incoming_hyperparameter if available
        hyperparameter = extract_parameters(full_hyperparameters, incoming_hyperparameter)

        create_info['config']['parameters'] = hyperparameter

        if parsed_args.max_epochs:
            create_info['config']['maxEpochs'] = int(parsed_args.max_epochs)

        create_info['config']['priority'] = 0
        if parsed_args.priority:
            create_info['config']['priority'] = float(parsed_args.priority)

        if parsed_args.max_time:
            create_info['config']['maxTime'] = float(parsed_args.max_time)

        if parsed_args.command:
            create_info['config']['command'] = parsed_args.command

        if parsed_args.image:

            # reset install options, since we can't make sure if the base image still fits
            if 'image' in config and config['image'] and config['image'] != parsed_args.image:
                create_info['config']['install'] = None

            # reset dockerfile, since we specified manually an image
            create_info['config']['dockerfile'] = None
            create_info['config']['image'] = parsed_args.image

        if parsed_args.server:
            create_info['config']['servers'] = []
            for name in parsed_args.server:
                create_info['config']['servers'].append(name)

        if 'resources' not in create_info['config']:
            create_info['config']['resources'] = {}

        if parsed_args.cpu or parsed_args.memory or parsed_args.gpu is not None or parsed_args.gpu_memory:
            if parsed_args.cpu: create_info['config']['resources']['cpu'] = float(parsed_args.cpu)
            if parsed_args.memory: create_info['config']['resources']['memory'] = float(parsed_args.memory)
            if parsed_args.gpu is not None: create_info['config']['resources']['gpu'] = float(parsed_args.gpu)
            if parsed_args.gpu_memory: create_info['config']['resources']['gpu_memory'] = float(parsed_args.gpu_memory)

        if parsed_args.local:
            # usually, the aetros server would assign resources at job root level from the assigned server
            # but since it's started locally, we just use the requested one. User should know what they do.
            # start.py will use 'config' stuff anyone for docker limitation, so we should make sure it is
            # being displayed.

            if 'image' in create_info['config'] and create_info['config']:
                resources = create_info['config']['resources']
                create_info['resources_assigned'] = {'cpus': 1, 'memory': 1, 'gpus': []}

                if 'gpu' in resources and resources['gpu'] > 0:
                    create_info['resources_assigned']['gpus'] = [1] * resources['gpu']
                if 'cpu' in resources:
                    create_info['resources_assigned']['cpus'] = resources['cpu']
                if 'memory' in resources:
                    create_info['resources_assigned']['memory'] = resources['memory']
            else:
                # since this runs on the host, extract machine hardware and put int resources_assigned
                # so we see it at the job.
                pass

        if parsed_args.local:
            create_info['server'] = 'local'

        create_info['config']['sourcesAttached'] = True

        if aetros.utils.git.get_current_commit_hash():
            create_info['origin_git_source'] = {
                'origin': aetros.utils.git.get_current_remote_url(),
                'author': aetros.utils.git.get_current_commit_author(),
                'message': aetros.utils.git.get_current_commit_message(),
                'branch': aetros.utils.git.get_current_branch(),
                'commit': aetros.utils.git.get_current_commit_hash(),
            }

        job.create(create_info=create_info, server=None)

        print("Job %s/%s created." % (job.model_name, job.job_id))

        if parsed_args.local:
            start(self.logger, job.model_name + '/' + job.job_id, fetch=False, env=env, volumes=parsed_args.volume, gpu_devices=parsed_args.gpu_device)
        else:
            if parsed_args.volume:
                print("Can not use volume with jobs on the cluster. Use datasets instead.")
                sys.exit(1)

            #todo, make it visible
            job.git.push()
            print("Open http://%s/model/%s/job/%s to monitor it." % (job.host, job.model_name, job.job_id))



from __future__ import absolute_import
from __future__ import print_function
import argparse
import subprocess

import sys

import os

from aetros.utils import read_home_config, setup_git_ssh


class PushJobCommand:
    def __init__(self, logger):
        self.logger = logger
        self.client = None
        self.registered = False
        self.active = True

    def main(self, args):
        import aetros.const

        parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter, prog=aetros.const.__prog__ + ' push-job')
        parser.add_argument('id', nargs='?', help="Model name like peter/mnist/ef8009d83a9892968097cec05b9467c685d45453")

        parsed_args = parser.parse_args(args)

        if not parsed_args.id:
            parser.print_help()
            sys.exit(1)

        config = read_home_config()

        model = parsed_args.id[0:parsed_args.id.rindex('/')]
        ref = 'refs/aetros/job/' + parsed_args.id[parsed_args.id.rindex('/')+1:]

        git_dir = os.path.normpath(config['storage_dir'] + '/' + model + '.git')

        if not os.path.isdir(git_dir):
            self.logger.error("Git repository for model %s in %s not found." % (parsed_args.id, git_dir))
            self.logger.error("Are you in the correct directory?")

        print('Pushing ' + ref + ' from ' + git_dir)
        setup_git_ssh(config)
        subprocess.call([config['git'], '--bare', '--git-dir', git_dir, 'push', 'origin', ref])